/**
 *  Copyright (C) 2013
 *
 *  RTCDD Implementation
 *
 *  This is the full implementation of RTC with both main server and secondary server.
 *  This version is the one used in all experimentations of RTC.
 *	The other version (rtc.cpp) is only used to measure the improvement of dependency detector.
 */

#include "../cm.hpp"
#include "algs.hpp"
#include "RedoRAWUtils.hpp"
#include <sys/mman.h>

// Don't just import everything from stm. This helps us find bugs.
using stm::TxThread;
using stm::timestamp;
using stm::WriteSetEntry;
using stm::ValueList;
using stm::ValueListEntry;
using stm::filter_t;
using stm::pad_word_t;

/*
 *  RTCDD definitions
 */

#define PAGE_SIZE				4096
#define CACHE_LINE_SIZE 		64

#define r_align(n, r)       	(((n) + (r) - 1) & -(r))
#define cache_align(n)       	r_align(n , CACHE_LINE_SIZE)
#define pad_to_cache_line(n) 	(cache_align(n) - (n))

// Server-Client Request states
#define READY 		0
#define PENDING		1
#define ABORTED		2

#define GET_NODES_CMD																										\
		"NODES=/sys/devices/system/node;"																			\
		"CPUS=/sys/devices/system/cpu;"																				\
		"if [ -d $NODES ]; then"																							\
	  "  for F in `ls -d $NODES/node*`; do if [ $(cat $F/cpulist | grep '-') ]; then seq -s ' ' $(cat $F/cpulist | tr -s '-' ' '); else cat $F/cpulist | tr -s ',' ' '; fi; done;" \
		"else"																																\
		"  find $CPUS -maxdepth 1 -name \"cpu[0-9]*\" -exec basename {} \\; | sed -e 's/cpu\\(.*\\)/\\1/g' | tr -s '\n' ' ';" \
		"fi"

/*
 *  RTCDD structures
 */
struct request {
	volatile int state; /* the state of commit request */
	TxThread* tx; 		/* Tx context */
	char pad[pad_to_cache_line(sizeof(int) + sizeof(TxThread*))];
};

/*
 * RTCDD Variables
 */

static struct request* requests = 0; /* requests array */
volatile unsigned int rtcdd_num_threads = 0;
volatile int rtcdd_num_cores = 0;

cpu_set_t rtcdd_client_cpuset;
cpu_set_t rtcdd_main_server_cpuset;
cpu_set_t rtcdd_secondary_server_cpuset;

volatile bool rtcdd_server_initialized = 0;
volatile pad_word_t parallel;
volatile pad_word_t main_server_curr_req_num;
volatile pad_word_t servers_lock;

namespace {

const uintptr_t VALIDATION_FAILED = 1;
NOINLINE uintptr_t validate_client(TxThread*);
NOINLINE uintptr_t validate_server(TxThread*);
bool irrevoc(STM_IRREVOC_SIG(,));
void onSwitchTo();

template<class CM>
struct RTCDD_Generic {
	static TM_FASTCALL bool begin(TxThread*);
	static TM_FASTCALL void commit(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_ro(STM_COMMIT_SIG(,));
	static TM_FASTCALL void commit_rw(STM_COMMIT_SIG(,));
	static TM_FASTCALL void* read_ro(STM_READ_SIG(,,));
	static TM_FASTCALL void* read_rw(STM_READ_SIG(,,));
	static TM_FASTCALL void write_ro(STM_WRITE_SIG(,,,));
	static TM_FASTCALL void write_rw(STM_WRITE_SIG(,,,));
	static stm::scope_t* rollback(STM_ROLLBACK_SIG(,,,));
	static void initialize(int id, const char* name);
};

/*
 * RTCDD Methods
 */

void* rcl_anon_mmap(size_t n) {
	void* res = mmap(0, n, PROT_READ | PROT_WRITE, MAP_ANONYMOUS | MAP_PRIVATE,
			-1, 0);
	if (res == MAP_FAILED)
		fprintf(stderr, "Cannot allocate anon mmap");
	return res;
}

static void rcl_extract_topology(const char* cmd_nodes) {
	FILE *file;
	char text[1024], *p, *saveptr;
	int nb_nodes = 0;
	int nb_cores = 0;

	/* We use UNIX commands to find the nodes. */
	if (!(file = popen(cmd_nodes, "r")))
		fprintf(stderr, "Cannot open system file");

	/* For each node	... */
	while (fgets(text, 1024, file)) {
		for (p = 0, p = strtok_r(text, " ", &saveptr); p;
				p = strtok_r(0, " ", &saveptr))
			nb_cores++;
		nb_nodes++;
	}

	printf("Number of nodes: %d\n", nb_nodes);
	printf("Number of cores: %d\n", nb_cores);

	rtcdd_num_cores = nb_cores;

	// close file
	if (pclose(file) < 0)
		fprintf(stderr, "Cannot close system file");

	// add first core to main server cpu-set, 
	// second core to secondary server cpu-set, 
	// and all other cores to clients cpu-set
	// These cpu-sets will be used in threads binding
	for (int i = 0; i < nb_cores; i++) {

		// first and second cores are dedicated for servers and other cores are dedicated for clients.
		if (i == 0)
			CPU_SET(i, &rtcdd_main_server_cpuset);
		else if (i == 1)
			CPU_SET(i, &rtcdd_secondary_server_cpuset);
		else
			CPU_SET(i, &rtcdd_client_cpuset);
	}
}

void* rcl_run_main_server(void* dummy) {

	// determine number of cores
	rcl_extract_topology(GET_NODES_CMD);

	// bind server thread to its core
	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
			&rtcdd_main_server_cpuset);
	printf("Main Server CPU is: %d\n", sched_getcpu());

	// Allocate "requests" and "bf_requests" Arrays
	size_t request_size =
			r_align(sizeof(struct request)*stm::MAX_THREADS, PAGE_SIZE);
	requests = (request*) rcl_anon_mmap(request_size);

	volatile unsigned int i = 0;
	for (i = 0; i < stm::MAX_THREADS; i++)
		requests[i].state = READY;

	// mark server as initialized so that clients can start
	main_server_curr_req_num.val = 0;
	rtcdd_server_initialized = true;

	request* req;
	volatile unsigned int curr_num_threads = rtcdd_num_threads;

	parallel.val = 0;
	servers_lock.val = 0;

	// server loop
	while (stm::clients_exist) {

		curr_num_threads = rtcdd_num_threads;

		for (i = 0; i < curr_num_threads; i++) {

			req = &(requests[i]);

			// Pending Transaction
			if (req->state == PENDING) {

				// Validation is needed because client may be invalidated after changing its state to PENDING
				if (timestamp.val != req->tx->start_time) {
					if ((req->tx->start_time = validate_server(req->tx)) == VALIDATION_FAILED) {
						req->state = ABORTED;
						continue;
					}
				}

				// Small write-set. Disable dependency detector
				if(req->tx->writes.size() < 20)
				{
					timestamp.val++;

					req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));
					CFENCE;

					timestamp.val++;
					req->state = READY;
				}
				// large write-set. Enable dependency detector
				else
				{
					parallel.val = 1;
					main_server_curr_req_num.val = i;
					timestamp.val++;

					req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));
					CFENCE;

					// Acquire servers lock before incrementing timestamp again to even number
					while(!bcasptr(&servers_lock.val, 0, 1));

					timestamp.val++;

					// release servers lock
					servers_lock.val = 0;
					req->state = READY;
					parallel.val = 0;
				}
			}
		}
	}

	// deallocate request array
	munmap(requests,request_size);

	// clear flag when killing server
	rtcdd_server_initialized = false;

	return NULL;
}

void* rcl_run_secondary_server(void* dummy) {

	// bind server thread to its core
	pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
			&rtcdd_secondary_server_cpuset);

	printf("Secondary Server CPU is: %d\n", sched_getcpu());

	request* req, *main_server_request;
	uintptr_t i = 0;
	volatile uintptr_t curr_num_threads = rtcdd_num_threads;

	volatile uintptr_t local_timestamp = 0;
	volatile uintptr_t local_main_server_req_num = 0;

	int count = 0;
	int aborted = 0;

	// server loop
	while (stm::clients_exist) {

		curr_num_threads = rtcdd_num_threads;

		for (i = 0; i < curr_num_threads; i++) {

			if(parallel.val == 0)
				continue;

			aborted = 0;
			local_timestamp = timestamp.val;
			local_main_server_req_num = main_server_curr_req_num.val;

			// Server is not executing a commit, or the same commit.
			if( (local_timestamp & 1) == 0 || local_main_server_req_num == i)
				continue;

			req = &(requests[i]);
			main_server_request = &(requests[local_main_server_req_num]);

			// pending request AND independent from main server
			if (req->state == PENDING && !(main_server_request->tx->wf->intersect(req->tx->rf))) {

				// validate read-set
				// Validation before acquiring servers lock, but changing the request state must be after acquiring servers lock
				if(local_timestamp != (req->tx->start_time + 1))
				{
					if ((req->tx->start_time = validate_server(req->tx)) == VALIDATION_FAILED)
						aborted = 1;
				}

				// Acquire servers lock
				if(!bcasptr(&servers_lock.val, 0, 1))
					continue;

				// check if main server still in the same point
				if(local_timestamp != timestamp.val)
				{
					servers_lock.val = 0;
					continue;
				}

				// Finalize commit/abort and then release the servers lock

				if(aborted)
				{
					req->state = ABORTED;
					servers_lock.val = 0;
				}
				else
				{
					req->tx->writes.writeback(STM_WHEN_PROTECT_STACK(upper_stack_bound));

					// Release the sequence lock and acknowledge the client
					CFENCE;
					req->state = READY;
					servers_lock.val = 0;
					while(local_timestamp == timestamp.val);
				}
				count++;
			}
		}
	}

	printf("Number of transactions in secondary server =%d\n", count);

	return NULL;
}

/*
 * End of RTCDD servers loops
 */

uintptr_t validate_client(TxThread* tx) {

	while (true) {

		// loop until no commit is executed
		uintptr_t s = timestamp.val;
        if ((s & 1) == 1)
            continue;

		// check the read set
		CFENCE;
		// don't branch in the loop---consider it backoff if we fail
		// validation early
		bool valid = true;
		foreach (ValueList, i, tx->vlist)
			valid &= i->isValid();

		if (!valid)
			return VALIDATION_FAILED;

		// restart if timestamp changed during read set iteration
		CFENCE;
		if (timestamp.val == s)
			return s;
	}
}

uintptr_t validate_server(TxThread* tx) {

	while (true) {

		// No need to loop until no commit is executed. Start immediately
		uintptr_t s = timestamp.val;

		// check the read set
		CFENCE;
		// don't branch in the loop---consider it backoff if we fail
		// validation early
		bool valid = true;
		foreach (ValueList, i, tx->vlist)
			valid &= i->isValid();

		if (!valid)
			return VALIDATION_FAILED;
		// restart if timestamp changed during read set iteration
		CFENCE;
		if (timestamp.val == s)
			return s;
	}
}

bool irrevoc(STM_IRREVOC_SIG(tx,upper_stack_bound)) {

	fprintf(stderr, "Irrevoc is not yet implemented");

	return true;
}

void onSwitchTo() {

	// We just need to be sure that the timestamp is not odd, or else we will
	// block.  For safety, increment the timestamp to make it even, in the event
	// that it is odd.

	if (timestamp.val & 1)
		++timestamp.val;

	// Create Server thread if not exist...
	if (!rtcdd_server_initialized) {
		void* args = (void*) (stm::MAX_THREADS);
		pthread_t tid;
		pthread_attr_t attr;
		pthread_attr_init(&attr);
		pthread_create(&tid, &attr, &rcl_run_main_server, args);

		// wait until server thread initializes its data-structures
		while (!rtcdd_server_initialized)
			;

		void* args2 = (void*) (stm::MAX_THREADS);
		pthread_t tid2;
		pthread_attr_t attr2;
		pthread_attr_init(&attr2);
		pthread_create(&tid2, &attr2, &rcl_run_secondary_server, args2);
	}
}

template<typename CM>
void RTCDD_Generic<CM>::initialize(int id, const char* name) {

	// set the name
	stm::stms[id].name = name;

	// set the pointers
	stm::stms[id].begin = RTCDD_Generic<CM>::begin;
	stm::stms[id].commit = RTCDD_Generic<CM>::commit_ro;
	stm::stms[id].read = RTCDD_Generic<CM>::read_ro;
	stm::stms[id].write = RTCDD_Generic<CM>::write_ro;
	stm::stms[id].irrevoc = irrevoc;
	stm::stms[id].switcher = onSwitchTo;
	stm::stms[id].privatization_safe = true;
	stm::stms[id].rollback = RTCDD_Generic<CM>::rollback;
}

template<class CM>
bool RTCDD_Generic<CM>::begin(TxThread* tx) {

	// bind thread to clinet core-set
	if (!tx->bounded) {
		pthread_setaffinity_np(pthread_self(), sizeof(cpu_set_t),
				&rtcdd_client_cpuset);

		//increase number of threads
		//TODO:: Do it more intelligently
		int curr_num_thread = rtcdd_num_threads;
		while (!bcasptr(&rtcdd_num_threads, curr_num_thread,
				curr_num_thread + 1))
			curr_num_thread = rtcdd_num_threads;

		tx->bounded = true;
	}

	// Sample the sequence lock, if it is even decrement by 1
	tx->start_time = timestamp.val & ~(1L);

	// refer to tx in request
	requests[tx->id - 1].tx = tx;

	// notify the allocator
	tx->allocator.onTxBegin();

	// notify CM
	CM::onBegin(tx);

	return false;
}

template<class CM>
void RTCDD_Generic<CM>::commit(STM_COMMIT_SIG(tx,upper_stack_bound)) {
	fprintf(stderr,
			"COMMIT IS NOT SUPPOSED TO BE CALLED. COMMIT_RO AND COMMIT_RW ARE ONLY USED");
}

template<class CM>
void RTCDD_Generic<CM>::commit_ro(STM_COMMIT_SIG(tx,)) {
	// Since all reads were consistent, and no writes were done, the read-only
	// NOrec transaction just resets itself and is done.
	CM::onCommit(tx);
	tx->vlist.reset();
	tx->rf->clear();
	OnReadOnlyCommit(tx);
}

template<class CM>
void RTCDD_Generic<CM>::commit_rw(STM_COMMIT_SIG(tx,upper_stack_bound)) {

	// Validation. Saves time of servers on useless invalidated requests
	if (timestamp.val != tx->start_time)
		if ((tx->start_time = validate_client(tx)) == VALIDATION_FAILED)
			tx->tmabort(tx);

	// send request
	requests[tx->id - 1].state = PENDING;
	volatile int local_state;

	// loop until server finished executing commit.
	do{
		local_state = requests[tx->id - 1].state;
	}while (local_state != READY && local_state != ABORTED)
		;

	if (local_state == ABORTED) {
		requests[tx->id - 1].state = READY;
		tx->tmabort(tx);
	} else {
		// Finalize commit

		// notify CM
		CM::onCommit(tx);

		tx->vlist.reset();
		tx->writes.reset();
		tx->rf->clear();
		tx->wf->clear();

		// This switches the thread back to RO mode.
		OnReadWriteCommit(tx, read_ro, write_ro, commit_ro);
	}
}

template<class CM>
void*
RTCDD_Generic<CM>::read_ro(STM_READ_SIG(tx,addr,mask)) {
	// A read is valid iff it occurs during a period where the seqlock does
	// not change and is even.  This code also polls for new changes that
	// might necessitate a validation.

	// read the location to a temp
	void* tmp = *addr;
	CFENCE;

	// if the timestamp has changed since the last read, we must validate and
	// restart this read
	while (tx->start_time != timestamp.val) {
		if ((tx->start_time = validate_client(tx)) == VALIDATION_FAILED)
			tx->tmabort(tx);

		tmp = *addr;
		CFENCE;
	}

	// log the address and value.
	tx->vlist.insert(STM_VALUE_LIST_ENTRY(addr, tmp, mask));
	// Update read-write filter (we used rf as the read-write filter here)
	tx->rf->add(addr);
	return tmp;
}

template<class CM>
void*
RTCDD_Generic<CM>::read_rw(STM_READ_SIG(tx,addr,mask)) {
	// check the log for a RAW hazard, we expect to miss
	WriteSetEntry log(STM_WRITE_SET_ENTRY(addr, NULL, mask));
	bool found = tx->writes.find(log);
	REDO_RAW_CHECK(found, log, mask);

	// Use the code from the read-only read barrier. This is complicated by
	// the fact that, when we are byte logging, we may have successfully read
	// some bytes from the write log (if we read them all then we wouldn't
	// make it here). In this case, we need to log the mask for the rest of the
	// bytes that we "actually" need, which is computed as bytes in mask but
	// not in log.mask. This is only correct because we know that a failed
	// find also reset the log.mask to 0 (that's part of the find interface).
	void* val = read_ro(tx, addr STM_MASK(mask & ~log.mask));
	REDO_RAW_CLEANUP(val, found, log, mask);
	return val;
}

template<class CM>
void RTCDD_Generic<CM>::write_ro(STM_WRITE_SIG(tx,addr,val,mask)) {
	// buffer the write, and switch to a writing context
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	// update write filter and read-write filter
	tx->wf->add(addr);
	tx->rf->add(addr);
	OnFirstWrite(tx, read_rw, write_rw, commit_rw);
}

template<class CM>
void RTCDD_Generic<CM>::write_rw(STM_WRITE_SIG(tx,addr,val,mask)) {
	// just buffer the write
	tx->writes.insert(WriteSetEntry(STM_WRITE_SET_ENTRY(addr, val, mask)));
	// update write filter and read-write filter
	tx->wf->add(addr);
	tx->rf->add(addr);
}

template<class CM>
stm::scope_t*
RTCDD_Generic<CM>::rollback(
		STM_ROLLBACK_SIG(tx, upper_stack_bound, except, len)) {

	stm::PreRollback(tx);

	// notify CM
	CM::onAbort(tx);

	// Perform writes to the exception object if there were any... taking the
	// branch overhead without concern because we're not worried about
	// rollback overheads.
	STM_ROLLBACK(tx->writes, upper_stack_bound, except, len);

    // reset filters and lists
	tx->vlist.reset();
	tx->rf->clear();
    if (tx->writes.size()) {
        tx->writes.reset();
        tx->wf->clear();
    }

	return stm::PostRollback(tx, read_ro, write_ro, commit_ro);

}
} // (anonymous namespace)

// Register RTCDD initializer functions. Do this as declaratively as
// possible. Remember that they need to be in the stm:: namespace.
#define FOREACH_RTCDD(MACRO)                    \
    MACRO(RTCDD, HyperAggressiveCM)             \
    MACRO(RTCDDHour, HourglassCM)               \
    MACRO(RTCDDBackoff, BackoffCM)              \
    MACRO(RTCDDHB, HourglassBackoffCM)

#define INIT_RTCDD(ID, CM)                      \
    template <>                                 \
    void initTM<ID>() {                         \
        RTCDD_Generic<CM>::initialize(ID, #ID);     \
    }

namespace stm {
FOREACH_RTCDD(INIT_RTCDD)
}

#undef FOREACH_RTCDD
#undef INIT_RTCDD
